/* eslint-disable no-magic-numbers */
import Polygon from 'ol/geom/Polygon';
import { Style } from 'ol/style';
import Feature, { FeatureLike } from 'ol/Feature';
import { MultiPolygon } from 'ol/geom';
import { UsePointToProjectionResult } from '@/utils/map/usePointToProjection';
import {
  HorizontalAlign,
  ScaleFunction,
  VerticalAlign,
} from '@/components/Map/MapViewer/MapViewerVector/MapViewerVector.types';
import { MapInstance } from '@/types/map';
import getTextStyle from '@/components/Map/MapViewer/MapViewerVector/utils/shapes/text/getTextStyle';
import { rgbToHex } from '@/components/Map/MapViewer/MapViewerVector/utils/shapes/style/rgbToHex';
import getTextCoords from '@/components/Map/MapViewer/MapViewerVector/utils/shapes/text/getTextCoords';
import { Color } from '@/types/models';
import {
  COMPLEX_SBO_TERM,
  MAP_ELEMENT_TYPES,
} from '@/components/Map/MapViewer/MapViewerVector/MapViewerVector.constants';
import VectorSource from 'ol/source/Vector';
import MapBackgroundsEnum from '@/redux/map/map.enums';
import { Extent } from 'ol/extent';
import { MapSize } from '@/redux/map/map.types';
import getCoverStyles from '@/components/Map/MapViewer/MapViewerVector/utils/shapes/style/getCoverStyles';
import handleSemanticView from '@/components/Map/MapViewer/MapViewerVector/utils/shapes/elements/handleSemanticView';

export interface BaseMapElementProps {
  type: string;
  sboTerm?: string;
  id: number;
  complexId?: number | null;
  compartmentId: number | null;
  x: number;
  y: number;
  width: number;
  height: number;
  zIndex: number;
  text: string;
  fontSize: number;
  nameX: number;
  nameY: number;
  nameWidth: number;
  nameHeight: number;
  fontColor: Color;
  nameVerticalAlign: VerticalAlign;
  nameHorizontalAlign: HorizontalAlign;
  fillColor: Color;
  borderColor: Color;
  pointToProjection: UsePointToProjectionResult;
  vectorSource: VectorSource;
  mapBackgroundType: number;
  mapSize: MapSize;
}

export default abstract class BaseMultiPolygon {
  type: string;

  sboTerm: string | undefined;

  id: number;

  complexId?: number | null;

  compartmentId: number | null;

  x: number;

  y: number;

  width: number;

  height: number;

  zIndex: number;

  text: string;

  fontSize: number;

  nameX: number;

  nameY: number;

  nameWidth: number;

  nameHeight: number;

  fontColor: Color;

  nameVerticalAlign: VerticalAlign;

  nameHorizontalAlign: HorizontalAlign;

  fillColor: Color;

  borderColor: Color;

  polygons: Array<Polygon> = [];

  styles: Array<Style> = [];

  feature: Feature = new Feature();

  pointToProjection: UsePointToProjectionResult;

  vectorSource: VectorSource;

  mapBackgroundType: number;

  mapSize: MapSize;

  mapExtentCache: Map<number, [number, number, number, number]> = new Map<
    number,
    [number, number, number, number]
  >();

  constructor({
    type,
    sboTerm,
    id,
    complexId,
    compartmentId,
    x,
    y,
    width,
    height,
    zIndex,
    text,
    fontSize,
    nameX,
    nameY,
    nameWidth,
    nameHeight,
    fontColor,
    nameVerticalAlign,
    nameHorizontalAlign,
    fillColor,
    borderColor,
    pointToProjection,
    vectorSource,
    mapBackgroundType,
    mapSize,
  }: BaseMapElementProps) {
    this.type = type;
    this.sboTerm = sboTerm;
    this.id = id;
    this.complexId = complexId;
    this.compartmentId = compartmentId;
    this.x = x;
    this.y = y;
    this.width = width;
    this.height = height;
    this.zIndex = zIndex;
    this.text = text;
    this.fontSize = fontSize;
    this.nameX = nameX;
    this.nameY = nameY;
    this.nameWidth = nameWidth;
    this.nameHeight = nameHeight;
    this.fontColor = fontColor;
    this.nameVerticalAlign = nameVerticalAlign;
    this.nameHorizontalAlign = nameHorizontalAlign;
    this.fillColor = fillColor;
    this.borderColor = borderColor;
    this.pointToProjection = pointToProjection;
    this.vectorSource = vectorSource;
    this.mapBackgroundType = mapBackgroundType;
    this.mapSize = mapSize;
  }

  protected abstract createPolygons(): void;

  protected drawText(): void {
    if (this.text) {
      const textCoords = getTextCoords({
        x: this.nameX,
        y: this.nameY,
        width: this.nameWidth,
        height: this.nameHeight,
        fontSize: this.fontSize,
        verticalAlign: this.nameVerticalAlign,
        horizontalAlign: this.nameHorizontalAlign,
        pointToProjection: this.pointToProjection,
      });
      const textPolygon = new Polygon([[textCoords, textCoords]]);
      textPolygon.set('type', MAP_ELEMENT_TYPES.TEXT);
      textPolygon.set('text', this.text);
      textPolygon.set('fontSize', this.fontSize);
      const textStyle = getTextStyle({
        text: this.text,
        fontSize: this.fontSize,
        color: rgbToHex(this.fontColor),
        zIndex: this.zIndex,
        horizontalAlign: this.nameHorizontalAlign,
      });
      textStyle.setGeometry(textPolygon);
      this.styles.push(textStyle);
      this.polygons.push(textPolygon);
    }
  }

  protected drawMultiPolygonFeature(mapInstance: MapInstance): void {
    this.feature = new Feature({
      geometry: new MultiPolygon(this.polygons),
      zIndex: this.zIndex,
      getScale: ((): ScaleFunction => {
        const maxZoom = mapInstance?.getView().get('originalMaxZoom');
        const minResolution = maxZoom
          ? mapInstance?.getView().getResolutionForZoom(maxZoom)
          : undefined;

        return (resolution: number): number => {
          if (minResolution) {
            return minResolution / resolution;
          }
          return 1;
        };
      })(),
      getMapExtent: (resolution: number): [number, number, number, number] | undefined => {
        if (this.mapExtentCache.has(resolution)) {
          return this.mapExtentCache.get(resolution);
        }

        const view = mapInstance?.getView();
        const center = view?.getCenter();
        const size = mapInstance?.getSize();

        if (!size || !center) {
          return undefined;
        }

        const extentWidth = size[0] * resolution;
        const extentHeight = size[1] * resolution;

        const extent: [number, number, number, number] = [
          center[0] - extentWidth / 2,
          center[1] - extentHeight / 2,
          center[0] + extentWidth / 2,
          center[1] + extentHeight / 2,
        ];

        this.mapExtentCache.set(resolution, extent);
        return extent;
      },
      id: this.id,
      complexId: this.complexId,
      compartmentId: this.compartmentId,
      type: this.type,
    });
    this.feature.setId(this.id);
    this.feature.setStyle(this.getStyle.bind(this));
  }

  protected getStyle(feature: FeatureLike, resolution: number): Style | Array<Style> | void {
    if (!(feature instanceof Feature)) {
      return undefined;
    }
    const styles: Array<Style> = [];
    const getScale = feature.get('getScale');
    let scale = 1;
    let cover = false;
    let largestExtent: Extent | null;

    if (getScale instanceof Function) {
      scale = getScale(resolution);
    }

    let hide = false;
    if (this.mapBackgroundType === MapBackgroundsEnum.SEMANTIC) {
      const semanticViewData = handleSemanticView(
        this.vectorSource,
        feature,
        resolution,
        scale,
        this.compartmentId,
        this.complexId,
      );
      cover = semanticViewData.cover;
      hide = semanticViewData.hide;
      largestExtent = semanticViewData.largestExtent;

      if (hide) {
        return undefined;
      }
    }

    let type: string;
    let fontSize: number;
    let lineWidth: number;
    let text: string;
    let coverStyle: Style | undefined;

    this.styles.forEach(style => {
      const styleGeometry = style.getGeometry();
      if (styleGeometry instanceof Polygon) {
        type = styleGeometry.get('type');
        text = styleGeometry.get('text');
        fontSize = styleGeometry.get('fontSize') || 10;
        lineWidth = styleGeometry.get('lineWidth');
        coverStyle = styleGeometry.get('coverStyle');
      }

      if (cover) {
        if (coverStyle && largestExtent) {
          styles.push(
            ...getCoverStyles(
              coverStyle,
              largestExtent,
              this.text,
              scale,
              this.zIndex + 1000,
              this.mapSize,
            ),
          );
        }
        return;
      }

      if (
        [MAP_ELEMENT_TYPES.MODIFICATION, MAP_ELEMENT_TYPES.TEXT].includes(type) &&
        scale * fontSize <= 4
      ) {
        return;
      }

      const clonedStyle = style.clone();
      const textStyle = clonedStyle.getText();
      const strokeStyle = clonedStyle.getStroke();
      if (type === 'text' && textStyle) {
        textStyle.setScale(scale);
        textStyle.setText(text);
      }
      if (strokeStyle && lineWidth) {
        if (lineWidth * scale < 0.08 && this.sboTerm !== COMPLEX_SBO_TERM) {
          clonedStyle.setStroke(null);
        } else {
          strokeStyle.setWidth(lineWidth * scale);
          const lineDash = strokeStyle.getLineDash();
          if (lineDash) {
            const newLineDash = lineDash.map(width => width * scale);
            strokeStyle.setLineDash(newLineDash);
          }
        }
      }
      styles.push(clonedStyle);
    });

    return styles;
  }
}
